import logging
import base64
import re
from json import JSONDecodeError
from typing import Optional, Any

import json
import asyncio

logger = logging.getLogger(__name__)


class PyOptimumException(Exception):
    """
    pyoptimum Exception
    """
    pass


class Client:
    """
    Client object to facilitate connection to the
    `optimize.vicbee.net <https:optimize.vicbee.net>`_ optimization api

    Calls will be made to a URL of the form:

    ``base_url/api/prefix/entry_point``

    in which ``entry_point`` is set in :meth:`pyoptimum.Client.call`.

    :param username: the username
    :param password: the password
    :param token: an authentication token
    :param auto_token_renewal: whether to automatically renew an expired token
    :param base_url: the api base url
    :param api: the target api
    :param prefix: the target api prefix
    :param auth_url: the api auth url (optional)
    """

    SLASH_END = re.compile('/+$')
    SLASH_START = re.compile('^/+')

    def __init__(self,
                 username: Optional[str] = None, password: Optional[str] = None,
                 token: Optional[str] = None,
                 auto_token_renewal: bool = True,
                 base_url: str = 'https://optimize.vicbee.net',
                 api: str = 'optimize',
                 prefix: str = 'api',
                 auth_url: str = ''):

        # username and password
        self.username = username
        self.password = password

        # token
        self.token = token

        # auto token renewal
        self.auto_token_renewal = auto_token_renewal

        if not self.token and (not self.username or not self.password):
            raise PyOptimumException("Token or username/password "
                                     "have not been provided.")

        # base url
        self.base_url = Client.url_join(base_url, api, prefix)

        # auth url
        self.auth_url = auth_url or self.base_url

        # initialize detail
        self.detail = None

    @staticmethod
    def url_join(*args: str) -> str:
        """
        Sanitizes and join url parts

        The main role of this function is to remove slashes at the beginning or end of url parts

        :param args: the parts of the url
        :return: the sanitized url
        """
        return '/'.join([x for x in
                         [Client.SLASH_START.sub('',
                                                 Client.SLASH_END.sub('', y))
                          for y in args] if x])

    def get_token(self) -> None:
        """
        Retrieve authentication token
        """
        from requests import get

        basic = base64.b64encode(bytes('{}:{}'.format(self.username, self.password),
                                       'utf-8')).decode('utf-8')
        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Basic ' + basic
        }

        response = get(Client.url_join(self.auth_url, 'get_token'), headers=headers)
        self.detail = None

        if response.ok:
            try:
                self.token = response.json().get('token')
            except (KeyError, JSONDecodeError) as e:
                raise PyOptimumException(f"Error while retrieving token: Invalid token: {e}")
            except Exception as e:
                raise PyOptimumException(f"Error while retrieving token: {e}")
        else:
            response.raise_for_status()

    def call(self, entry_point: str, data: Any) -> Any:
        """
        Calls the api ``entry_point`` with ``data``

        :param entry_point: the api entry point
        :param data: the data
        :return: dictionary with the response
        """

        from requests import post

        if self.token is None and not self.auto_token_renewal:
            raise PyOptimumException('No token available. Call get_token first')

        elif self.auto_token_renewal:
            # try renewing token
            self.get_token()

        # See https://github.com/psf/requests/issues/6014
        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
            'X-Api-Key': self.token
        }
        response = post(Client.url_join(self.base_url, entry_point),
                        data=json.dumps(data),
                        headers=headers)

        if response.ok:

            try:
                self.detail = None
                return response.json()
            except JSONDecodeError as e:
                raise PyOptimumException(f"Invalid response: {e}")

        else:

            if response.status_code == 400:
                content = json.loads(response.content)
                self.detail = content.get('detail', None)
                if self.detail:
                    raise PyOptimumException(self.detail)
            response.raise_for_status()

class AsyncClient(Client):
    """
    Async client object to facilitate connection to the
    `optimize.vicbee.net <https:optimize.vicbee.net>`_ optimization api

    Calls will be made to a URL of the form:

    ``base_url/api/prefix/entry_point``

    in which ``entry_point`` is set in :meth:`pyoptimum.Client.call`.

    :param username: the username
    :param password: the password
    :param token: an authentication token
    :param auto_token_renewal: whether to automatically renew an expired token
    :param base_url: the api base url
    :param api: the target api
    :param prefix: the target api prefix
    :param auth_url: the api auth url (optional)
    """

    async def get_token(self) -> None:
        """
        Retrieve authentication token
        """
        from aiohttp import ClientSession

        basic = base64.b64encode(bytes('{}:{}'.format(self.username, self.password),
                                       'utf-8')).decode('utf-8')
        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Basic ' + basic
        }

        self.detail = None
        async with ClientSession() as session:
            async with session.get(Client.url_join(self.auth_url, 'get_token'),
                                   headers=headers, raise_for_status=True) as resp:
                if resp.status < 400:
                    try:
                        response = await resp.json()
                        self.token = response.get('token')
                    except JSONDecodeError as e:
                        raise PyOptimumException(f"Invalid response: {e}")
                else:
                    resp.raise_for_status()


    async def call(self, entry_point: str, data: Any,
                   follow_resource: bool=False,
                   wait_time: float=1, max_retries: int=10) -> Any:
        """
        Calls the api ``entry_point`` with ``data``

        :param entry_point: the api entry point
        :param data: the data
        :return: dictionary with the response
        :param follow_resource: whether to automatically retrieve resource if calculation is deferred
        :param wait_time: how many seconds to wait before pooling resource again
        :param max_retries: maximum number of retries
        """

        from aiohttp import ClientSession

        if self.token is None and not self.auto_token_renewal:
            raise PyOptimumException('No token available. Call get_token first')

        elif self.auto_token_renewal:
            # try renewing token
            await self.get_token()

        # See https://github.com/psf/requests/issues/6014
        headers = {
            'Content-type': 'application/json',
            'Accept': 'application/json',
            'X-Api-Key': self.token
        }
        async with ClientSession() as session:
            async with session.post(Client.url_join(self.base_url, entry_point),
                                    data=json.dumps(data),
                                    headers=headers) as resp:
                if resp.status < 400:

                    # retrieve data
                    try:
                        self.detail = None
                        data = await resp.json()
                    except JSONDecodeError as e:
                        raise PyOptimumException(f"Invalid response: {e}")

                    if resp.status == 202 and follow_resource:
                        # calculation is deferred

                        # pool resource for status
                        id = data['id']
                        logger.debug("will wait for resource '%s'", id)
                        k = 0
                        while resp.status != 302:

                            # sleep
                            logger.debug('will sleep (k = %d)', k)
                            await asyncio.sleep(32 * (1 - 0.97 ** k) + 1)

                            # pool resource
                            logger.debug('pooling resource')
                            async with session.get(Client.url_join(self.base_url,
                                                                   f'resource/{id}/status'),
                                                   data=json.dumps(data),
                                                   headers=headers,
                                                   allow_redirects=False) as resp:
                                logger.debug(f'status = {resp.status}')
                                if resp.status == 302:
                                    # resource is ready
                                    logger.debug('resource is ready')
                                    break
                                elif resp.status >= 400:
                                    # raise exception
                                    content = await resp.json()
                                    self.detail = content.get('detail', None)
                                    if self.detail:
                                        raise PyOptimumException(self.detail)
                                elif resp.status == 200:
                                    content = await resp.json()
                                    logger.debug("status = '%s'", content)

                                k += 1
                                if k > max_retries:
                                    self.detail = 'Maximum number of retries exceeded'
                                    if self.detail:
                                        raise PyOptimumException(self.detail)

                        # resource is ready, retrieve value
                        logger.debug('retrieving resource')
                        async with session.get(Client.url_join(self.base_url,
                                                               f'resource/{id}/value'),
                                               data=json.dumps(data),
                                               headers=headers) as resp:
                            if resp.status >= 400:

                                # raise exception
                                content = await resp.json()
                                self.detail = content.get('detail', None)
                                if self.detail:
                                    raise PyOptimumException(self.detail)

                            else:

                                # retrieve data
                                logger.debug('got data')
                                try:
                                    self.detail = None
                                    data = await resp.json()
                                except JSONDecodeError as e:
                                    raise PyOptimumException(f"Invalid response: {e}")

                    return data

                elif resp.status == 400:
                    content = json.loads(await resp.content.read())
                    self.detail = content.get('detail', None)
                    if self.detail:
                        raise PyOptimumException(self.detail)

                resp.raise_for_status()



import io
import datetime
from pathlib import Path
from typing import Optional, Literal, Any, Union, List, Tuple, Dict

import numpy as np
import pandas as pd

from pyoptimum import AsyncClient
from pyoptimum.model import Model

LESS_THAN_OR_EQUAL = "\u2264"
GREATER_THAN_OR_EQUAL = "\u2265"

class Portfolio:

    ModelMethodLiteral = Literal['linear', 'linear-fractional']

    MethodLiteral = Literal['approximate', 'optimal', 'none']
    ConstraintFunctionLiteral = Literal['purchases', 'sales', 'holdings', 'short sales']
    ConstraintSignLiteral = Literal[LESS_THAN_OR_EQUAL, GREATER_THAN_OR_EQUAL]
    ConstraintUnitLiteral = Literal['shares', 'value', 'percent value']

    def __init__(self,
                 portfolio_client: AsyncClient,
                 model_client: AsyncClient,
                 model_method: ModelMethodLiteral = 'linear'):
        self.portfolio_client = portfolio_client
        self.model_client = model_client
        self.model_method = model_method
        self.models: dict = {}
        self.model_weights: Dict[str, float] = {}
        self.portfolio = None
        self.frontier = None
        self.frontier_query_params = {}
        self.frontier_method: Portfolio.MethodLiteral = 'none'

    @staticmethod
    def _locate_value(value: Any, column: str, df: pd.DataFrame) -> Tuple[Optional[pd.Series], Optional[pd.Series]]:
        index = df[column].searchsorted(value)
        n = df.shape[0]
        if index == n:
            # got the last element
            return df.iloc[-1], None
        elif index == 0:
            # got the first element
            return None, df.iloc[0]
        else:
            # interpolate
            return df.iloc[index - 1], df.iloc[index]

    def _update_prices(self, prices: dict) -> float:

        # add prices to dataframe
        prices = pd.DataFrame.from_dict(prices, orient='index',
                                        columns=['timestamp', 'close'])

        # update portfolio value and weights
        self.portfolio['close ($)'] = prices['close']
        self.portfolio['value ($)'] = self.portfolio['shares'] * self.portfolio['close ($)']
        value = self.get_value()
        self.portfolio['value (%)'] = self.portfolio['value ($)'] / value if value > 0.0 else 0.0

        return value

    def _get_portfolio_query(self,
                             cashflow: float, max_sales: float,
                             short_sales: bool, buy: bool, sell: bool,
                             rho: float=0.0) -> dict:

        # get model data
        model = self.get_model()
        data = model.to_dict(('r', 'Q', 'D', 'F'),
                             as_list=True,
                             normalize=True)

        # has regularization
        if rho > 0:
            data['rho'] = rho

        # get portfolio data
        value0 = self.get_value()
        value = value0 + cashflow
        x0 = self.portfolio['value (%)']
        data['x0'] = x0.tolist()

        # check bound in max_sales
        if np.isfinite(max_sales) and max_sales < 0:
            raise ValueError('max_sales must be non-negative')

        # add cashflow and constraints
        if value0 > 0.0:
            data['cashflow'] = cashflow / value0
            if np.isfinite(max_sales):
                data['constraints'] = [
                    {'label': 'sales', 'function': 'sales', 'bounds': max_sales / value0}]
        else:
            if cashflow == 0:
                raise ValueError("Cashflow cannot be zero on a portfolio with no shares")
            if np.isfinite(max_sales) and not short_sales:
                sell = False

            data['cashflow'] = 1

        data['options'] = {
            'short': short_sales,
            'buy': buy,
            'sell': sell
        }

        # has lower bound
        if np.isfinite(self.portfolio['lower']).any():
            xlo = self.portfolio['lower'] * self.portfolio['close ($)'] / value
            data['xlo'] = xlo.tolist()

        # has upper bound
        if np.isfinite(self.portfolio['upper']).any():
            xup = self.portfolio['upper'] * self.portfolio['close ($)'] / value
            data['xup'] = xup.tolist()

        return data

    def invalidate_model(self):
        """
        Invalidate the current portfolio models
        """
        self.models = {}
        self.model_weights = {}

    def invalidate_frontier(self):
        """
        Invalidate the current frontier
        """
        self.frontier = None
        self.frontier_query_params = {}
        self.frontier_method = 'none'

    def has_prices(self) -> bool:
        """
        :return: True if prices have been retrieved
        """
        return self.portfolio is not None and 'close ($)' in self.portfolio

    def has_frontier(self) -> bool:
        """
        :return: True if a frontier is available
        """
        return self.frontier is not None

    def has_models(self) -> bool:
        """
        :return: True if models have been retrieved
        """
        return bool(self.models)

    def set_models(self, models: Dict[str, Union[dict, Model]],
                   model_weights: Optional[Dict[str, float]]=None) -> None:
        """
        Set portfolio models

        :param models: a dictionary with the models per range
        :param model_weights: the model weights
        """
        # add models
        self.models = {rg: Model(data) for rg, data in models.items()}

        # set model weights
        model_weights = model_weights or {rg: 1.0 for rg in models.keys()}
        self.set_models_weights(model_weights)

        # reinitialize frontier
        self.invalidate_frontier()

    def get_model(self) -> Model:
        """
        :return: the portfolio model for the current model and weights
        """
        assert self.has_models(), "Models have not yet been retrieved"

        if self.model_method == 'linear' or len(self.models) == 1:
            # linear model
            model = Model({attr: sum([weight * getattr(self.models[rg], attr)
                                      for rg, weight in self.model_weights.items()])
                           for attr in ['r', 'D', 'F', 'Q']})
        else:
            # linear-fractional model
            model = Model({attr: sum([weight * getattr(self.models[rg], attr)
                                      for rg, weight in self.model_weights.items()])
                           for attr in ['r', 'Di', 'F', 'Q']})

        return model

    def get_tickers(self) -> List[str]:
        """
        :return: the portfolio tickers
        """
        return self.portfolio.index.tolist()

    def get_value(self) -> float:
        try:
            return sum(self.portfolio['value ($)'])
        except (KeyError, TypeError):
            return 0.0

    def import_csv(self, filepath: Union[str,bytes,io.BytesIO,Path]) -> None:
        """
        Import portfolio from csv file

        :param filepath: the file path
        """
        # read csv
        portfolio = pd.read_csv(filepath)

        # sanitize column names
        portfolio.columns = [c.strip() for c in portfolio.columns]

        # set ticker as index
        assert 'ticker' in portfolio.columns, "Portfolio must contain a column " \
                                              "named 'ticker'"
        portfolio.set_index('ticker', inplace=True)

        # initialize shares
        if 'shares' not in portfolio.columns:
            portfolio['shares'] = 0.0

        # make sure shares are float
        portfolio = portfolio.astype({"shares": float})

        # initialize lower and upper
        portfolio['lower'] = -np.inf
        portfolio['upper'] = np.inf

        # invalidate models and frontier
        self.invalidate_frontier()
        self.invalidate_model()

        # remove all other columns and set portfolio
        self.portfolio = portfolio[['shares', 'lower', 'upper']]

    async def retrieve_prices(self) -> float:
        """
        Retrieve the prices of the portfolio

        :return: the total portfolio value
        """

        # retrieve prices
        data = {'symbols': self.portfolio.index.tolist()}
        prices = await self.model_client.call('prices', data,
                                              follow_resource=True)

        # add prices to dataframe
        value = self._update_prices(prices)

        return value

    async def retrieve_models(self,
                              market_tickers: List[str],
                              ranges: Union[str, List[str]],
                              end: datetime.date = datetime.date.today(),
                              common_factors: bool = False,
                              include_prices: bool = False,
                              model_weights: Dict[str, float] = None) -> List[str]:
        """
        Retrieve portfolio models based on market tickers

        :param market_tickers: the market tickers
        :param ranges: the ranges to retrieve the portfolio models
        :param end: the last day to retrieve models
        :param common_factors: whether to keep factors common
        :param include_prices: whether to include prices on results
        :param model_weights: the model weights
        :return a list of messages
        """

        # retrieve models
        data = {
            'tickers': self.portfolio.index.tolist(),
            'market': market_tickers,
            'end': str(end),
            'range': ranges,
            'options': {
                'common': common_factors,
                'include_prices': include_prices
            }
        }
        models = await self.model_client.call('model', data,
                                              follow_resource=True)

        if include_prices:
            # update prices
            self._update_prices(models.pop('prices'))

        # remove messages
        messages = models.pop('messages')

        # set models
        self.set_models(models, model_weights)

        return messages

    async def retrieve_frontier(self,
                                cashflow: float, max_sales: float,
                                short_sales: bool, buy: bool, sell: bool,
                                rho: float=0.0) -> None:
        """
        Retrieve the portfolio frontier

        :param cashflow: the cashflow
        :param max_sales: the max sales
        :param short_sales: whether to allow short sales
        :param buy: whether to allow buys
        :param sell: whether to allow sells
        :param rho: regularization factor
        """

        # assert models and prices are defined
        assert self.has_prices() and self.has_models(),\
            "Either prices or models are missing"

        # retrieve frontier
        query = self._get_portfolio_query(cashflow, max_sales,
                                          short_sales, buy, sell, rho)
        sol = await self.portfolio_client.call('frontier', query,
                                               follow_resource=True)

        if len(sol['frontier']) == 0:
            self.invalidate_frontier()
            raise ValueError('Could not calculate optimal frontier; constraints likely make the problem infeasible.')

        # calculate variance
        model = self.get_model()
        values = []
        for s in sol['frontier']:
            if s['sol']['status'] == 'optimal':
                mu = s['mu']
                x = np.array(s['sol']['x'])
                _, std = model.return_and_variance(x)
                values.append((mu, std, x))

        # assemble return dataframe
        frontier = pd.DataFrame(values, columns=['mu', 'std', 'x'])
        self.frontier = frontier

        # save query params
        self.frontier_query_params = query
        self.frontier_method = 'optimal'

    async def retrieve_recommendation(self, mu: Optional[float]=None,
                                      method: MethodLiteral = 'approximate') -> dict:
        """
        Retrieve or calculate recommendations

        :param mu: the expected return
        :param method: whether to calculate approximate recommendations or retrieve from API
        :return:
        """

        assert self.has_frontier(), "Frontier has not been retrieved"

        if mu is None:
            # locate std
            _, std = self.get_return_and_variance()
            left, right = Portfolio._locate_value(std, 'std', self.frontier)
            if left is None:
                # got the first element
                mu = right['mu']
            elif right is None:
                # got the last element
                mu = left['mu']
            else:
                # interpolate
                std1, std2 = left['std'], right['std']
                eta = (std - std1)/(std2-std1)
                mu = (1 - eta) * left['mu'] + eta * right['mu']

        if method == 'approximate':
            # calculate approximate weights

            # locate mu
            left, right = Portfolio._locate_value(mu, 'mu', self.frontier)
            if left is None:
                # got the first element
                x = right['x']
                std = right['std']
            elif right is None:
                # got the last element
                x = left['x']
                std = left['std']
            else:
                # interpolate
                mu1, mu2 = left['mu'], right['mu']
                eta = (mu - mu1)/(mu2-mu1)
                x = (1 - eta) * left['x'] + eta * right['x']
                std = (1 - eta) * left['std'] + eta * right['std']

            return {'x': x, 'status': 'optimal', 'std': std, 'mu': mu}

        elif method == 'exact':
            # exact recommendation
            data = self.frontier_query_params.copy()
            data['mu'] = mu

            recs = await self.portfolio_client.call('portfolio', data,
                                                    follow_resource=True)
            if recs['status'] == 'optimal':
                x = np.array(recs['x'])
                _, std = self.get_model().return_and_variance(x)
                return {'x': x, 'status': recs['status'], 'std': std, 'mu': mu}
            else:
                # return approximate if optimization failed
                return await self.retrieve_recommendation(mu, method='approximate')

    def set_models_weights(self, model_weights: Dict[str, float]) -> None:
        """
        Set weights for the current portfolio models

        :param model_weights: the model weights
        """

        # range checking
        assert self.has_models(), "Models have not yet been retrieved"

        assert self.models.keys() == model_weights.keys(), ("Weights must have the same "
                                                            "keys as models")

        assert all(value >= 0 for value in model_weights.values()), ("Weights must be "
                                                                     "non-negative")

        # normalize weights
        sum_model_weights = sum(model_weights.values())
        if sum_model_weights > 0:
            self.model_weights = {rg: value/sum_model_weights for rg, value in
                                  model_weights.items()}
        else:
            self.model_weights = {rg: 1/len(model_weights) for rg in
                                  model_weights.keys()}

        # update frontier
        if self.has_frontier():
            model = self.get_model()
            mu_std = (self.frontier['x']
                      .apply(lambda x: pd.Series(model.return_and_variance(x),
                                                 index=['mu', 'std'])))
            self.frontier[['mu', 'std']] = mu_std
            self.frontier_method = 'approximate'

    def set_model_method(self, method: ModelMethodLiteral):
        self.model_method = method

    def get_portfolio_dataframe(self):
        """
        :return: the portfolio as a dataframe
        """

        # copy portfolio
        portfolio_df = self.portfolio.copy()

        if self.has_models():
            # add return
            model = self.get_model()
            portfolio_df['return (%)'] = model.r.copy()
            portfolio_df['std (%)'] = model.std.copy()

        if self.has_prices():
            # add lower and upper bounds in value
            portfolio_df['lower ($)'] = portfolio_df['close ($)'] * portfolio_df['lower']
            portfolio_df['upper ($)'] = portfolio_df['close ($)'] * portfolio_df['upper']

        return portfolio_df

    def get_frontier_range(self):
        """
        :return: the range of frontier values
        """
        if self.frontier is None:
            raise ValueError('Frontier has not been retrieved yet')
        mu_range = self.frontier['mu'].iloc[0], self.frontier['mu'].iloc[-1]
        std_range = self.frontier['std'].iloc[0], self.frontier['std'].iloc[-1]
        return mu_range, std_range

    def get_range(self) -> Tuple[List[float], List[float]]:
        """
        :return: the range of the current model
        """

        model = self.get_model()
        mu_range = [model.r.min(), model.r.max()]
        std_range = [model.std.min(), model.std.max()]

        if self.frontier is not None:
            # account for frontier
            fmu_range, fstd_range = self.get_frontier_range()
            mu_range = min(mu_range[0], fmu_range[0]), max(mu_range[1], fmu_range[1])
            std_range = min(std_range[0], fstd_range[0]), max(std_range[1], fstd_range[1])

        return mu_range, std_range

    def get_return_and_variance(self) -> Tuple[float, float]:
        """
        :return: the return and standard deviation of the current portfolio
        """
        return self.get_model().return_and_variance(self.portfolio['value (%)'])

    def get_unconstrained_frontier(self, x_bar: float=1.):
        """
        :return: the unconstrained frontier parameters
        """
        return self.get_model().unconstrained_frontier(x_bar)

    def remove_constraints(self, tickers: List[str]) -> None:
        """
        Remove all individual constraints on the listed tickers

        :param tickers: the list of tickers
        """

        # quick return
        if not tickers:
            return

        self.portfolio.loc[tickers, 'lower'] = -np.inf
        self.portfolio.loc[tickers, 'upper'] = np.inf

    def apply_constraint(self, tickers: List[str],
                         function: ConstraintFunctionLiteral,
                         sign: ConstraintSignLiteral,
                         value: Union[List[float], float],
                         unit: ConstraintUnitLiteral,
                         short_sales: bool=True, buy: bool=True, sell: bool=True) -> None:
        """
        Apply constraints to list of tickers

        :param tickers: the list of tickers
        :param function: the function to apply on the left-hand side of the inequality
        :param sign: the sign of the inequality
        :param value: the value of the right-hand side of the inequality
        :param unit: the unit in which the constraint is expressed
        :param short_sales: whether to allow short sales
        :param buy: whether to allow buying
        :param sell: whether to allow selling
        """
        # quick return
        if not tickers:
            return

        # make sure value is array
        if isinstance(value, (int, float)):
            value = [value] * len(tickers)
        value = np.array(value, dtype='float64')

        # make sure value is in shares
        shares = self.portfolio.loc[tickers, 'shares'].values
        if unit == 'value':
            close = self.portfolio.loc[tickers, 'close ($)'].values
            value /= close
        elif unit == 'percent value':
            value *= shares / 100

        # initialize
        lb, ub = None, None

        # sales
        if function == 'sales':
            if sign == LESS_THAN_OR_EQUAL:
                lb = shares - value
            elif sign == GREATER_THAN_OR_EQUAL:
                ub = shares - value
        elif function == 'purchases':
            if sign == LESS_THAN_OR_EQUAL:
                ub = shares + value
            elif sign == GREATER_THAN_OR_EQUAL:
                lb = shares + value
        elif function == 'short sales':
            if sign == LESS_THAN_OR_EQUAL:
                lb = -value
            elif sign == GREATER_THAN_OR_EQUAL:
                ub = -value
        elif function == 'holdings':
            if sign == LESS_THAN_OR_EQUAL:
                ub = value
            elif sign == GREATER_THAN_OR_EQUAL:
                lb = value

        # short sales
        if not short_sales:
            if lb is not None:
                lb = np.where(lb > 0, lb, 0)
            if ub is not None:
                ub = np.where(ub > 0, ub, 0)

        # no buys
        if not buy:
            if lb is not None:
                lb = np.where(lb > shares, shares, lb)
            if ub is not None:
                ub = np.where(ub > shares, shares, ub)

        # no sells
        if not sell:
            if lb is not None:
                lb = np.where(lb < shares, shares, lb)
            if ub is not None:
                ub = np.where(ub < shares, shares, ub)

        # apply bounds
        if lb is not None:
            self.portfolio.loc[tickers, 'lower'] = np.maximum(self.portfolio.loc[tickers, 'lower'], lb)

        if ub is not None:
            self.portfolio.loc[tickers, 'upper'] = np.minimum(self.portfolio.loc[tickers, 'upper'], ub)



import unittest
import os

import aiohttp
import math


from src import pyoptimum

username = 'demo@optimize.vicbee.net'
password = 'optimize'
base_url = os.getenv('TEST_BASE_URL', 'https://optimize.vicbee.net')

class TestBasic(unittest.IsolatedAsyncioTestCase):

    def test_constructor(self):

        client = pyoptimum.AsyncClient(username=username, password=password)
        self.assertIsNone(client.token)

        client = pyoptimum.AsyncClient(token='token')
        self.assertIsNone(client.username)
        self.assertIsNone(client.password)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.AsyncClient, username=username)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.AsyncClient, username=password)

        self.assertRaises(pyoptimum.PyOptimumException,
                          pyoptimum.AsyncClient, token='')

    async def test_optimize(self):

        client = pyoptimum.AsyncClient(username=username, password=password,
                                  base_url=base_url)

        self.assertIsNone(client.token)
        await client.get_token()
        self.assertIsNotNone(client.token)

        # wrong password
        client = pyoptimum.AsyncClient(username=username, password=password + 'wrong')

        self.assertIsNone(client.token)
        with self.assertRaises(aiohttp.client_exceptions.ClientResponseError):
            await client.get_token()

        # wrong constructor
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.AsyncClient)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.AsyncClient,
                          username=username)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.AsyncClient,
                          password=password)
        self.assertRaises(pyoptimum.PyOptimumException, pyoptimum.AsyncClient,
                          token='')

    async def test_models(self):

        client = pyoptimum.AsyncClient(username=username, password=password,
                                       base_url=base_url, api='models')

        self.assertIsNone(client.token)
        await client.get_token()
        self.assertIsNotNone(client.token)

    async def test_portfolio(self):

        client = pyoptimum.AsyncClient(username=username, password=password,
                                       base_url=base_url)

        s1 = 0.06
        s2 = 0.03
        rho = 1
        data = {
            'Q': [[s1 ** 2, s1 * s2 * rho], [s1 * s2 * rho, s2 ** 2]],
            'cashflow': 1,
            'mu': 0.11,
            'r': [.14, .08],
        }
        response = await client.call('portfolio', data)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        status = response.get('status')
        self.assertEqual(status, 'optimal')

        x = response.get('x')
        self.assertAlmostEqual(x[0], .5)
        self.assertAlmostEqual(x[1], .5)

        self.assertIsNone(client.detail)

        # call with slash
        response = await client.call('/portfolio', data)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        # call with errors
        data['r'] = [.14, .08, 0]
        with self.assertRaises(pyoptimum.PyOptimumException):
            await client.call('portfolio', data)
        self.assertIn('must be an array', client.detail)

    async def test_portfolio_async(self):

        client = pyoptimum.AsyncClient(username=username, password=password,
                                       base_url=base_url)

        s1 = 0.06
        s2 = 0.03
        rho = 1
        data = {
            'Q': [[s1 ** 2, s1 * s2 * rho], [s1 * s2 * rho, s2 ** 2]],
            'cashflow': 1,
            'mu': 0.11,
            'r': [.14, .08],
            'is_async': True
        }
        response = await client.call('portfolio', data,
                                     follow_resource=True, wait_time=1)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        status = response.get('status')
        self.assertEqual(status, 'optimal')

        x = response.get('x')
        self.assertAlmostEqual(x[0], .5)
        self.assertAlmostEqual(x[1], .5)

        self.assertIsNone(client.detail)

        # call with slash
        response = await client.call('/portfolio', data,
                                     follow_resource=True)

        obj = response.get('obj')
        self.assertTrue(math.fabs(math.sqrt(obj) - .045) < 1e-5)

        # call with errors
        data['r'] = [.14, .08, 0]
        with self.assertRaises(pyoptimum.PyOptimumException):
            await client.call('portfolio', data,
                              follow_resource=True)
        self.assertIn('must be an array', client.detail)

        # call with errors
        data['r'] = [.14, .08]
        data['options'] = { 'mu_max': -1 }
        with self.assertRaises(pyoptimum.PyOptimumException):
            await client.call('frontier', data,
                              follow_resource=True)
        self.assertIn('is larger than mu_max', client.detail)

    async def test_forbidden(self):

        client = pyoptimum.AsyncClient(username=username, password=password,
                                  base_url=base_url)

        data = {
            'A': [[2]],
            'blo': [0],
            'bup': [4],
            'c': [-1],
            'xlo': [0],
            'xup': [2]
        }
        with self.assertRaises(aiohttp.client_exceptions.ClientResponseError):
            await client.call('lp', data)



import typing
import unittest
import os

import numpy as np

username = 'demo@optimize.vicbee.net'
password = 'optimize'
base_url = os.getenv('TEST_BASE_URL', 'https://optimize.vicbee.net')


class TestBasic(unittest.TestCase):

    def setUp(self):

        import pyoptimum

        self.portfolio_client = pyoptimum.AsyncClient(username=username, password=password,api='optimize')
        self.model_client = pyoptimum.AsyncClient(username=username, password=password,api='models')

    def test_constructor(self):

        from pyoptimum.portfolio import Portfolio

        portfolio = Portfolio(self.portfolio_client, self.model_client)
        self.assertIsInstance(portfolio, Portfolio)
        self.assertFalse(portfolio.has_models())
        self.assertFalse(portfolio.has_frontier())

        self.assertEqual(portfolio.get_value(), 0.0)

        from pathlib import Path
        file = Path(__file__).parent / 'test.csv'
        portfolio.import_csv(file)
        self.assertListEqual(portfolio.portfolio.columns.tolist(),['shares', 'lower', 'upper'])
        self.assertListEqual(portfolio.portfolio.index.tolist(),['AAPL', 'MSFT', 'ASML', 'TQQQ'])
        self.assertListEqual(portfolio.portfolio['shares'].tolist(), [1, 10, 0, 13])

        self.assertEqual(portfolio.get_value(), 0.0)


class TestPortfolio(unittest.IsolatedAsyncioTestCase):

    def setUp(self):

        import pyoptimum
        from pyoptimum.portfolio import Portfolio

        self.portfolio_client = pyoptimum.AsyncClient(username=username, password=password,api='optimize')
        self.model_client = pyoptimum.AsyncClient(username=username, password=password,api='models')
        self.portfolio = Portfolio(self.portfolio_client, self.model_client)
        from pathlib import Path
        file = Path(__file__).parent / 'test.csv'
        self.portfolio.import_csv(file)

    async def test_prices(self):

        self.assertEqual(self.portfolio.get_value(), 0.0)

        # retrieve prices
        self.assertFalse(self.portfolio.has_prices())
        await self.portfolio.retrieve_prices()
        self.assertTrue(self.portfolio.has_prices())

        self.assertEqual(self.portfolio.get_value(), sum(self.portfolio.portfolio['value ($)']))

        self.assertIn('close ($)', self.portfolio.portfolio)
        self.assertIn('value ($)', self.portfolio.portfolio)
        self.assertIn('value (%)', self.portfolio.portfolio)
        np.testing.assert_array_equal(self.portfolio.portfolio['value ($)'], self.portfolio.portfolio['close ($)'] * self.portfolio.portfolio['shares'])
        np.testing.assert_array_equal(self.portfolio.portfolio['value (%)'], self.portfolio.portfolio['value ($)'] / sum(self.portfolio.portfolio['value ($)']))

        with self.assertRaises(AssertionError):
            await self.portfolio.retrieve_frontier(0, 0, False, True, True)

    async def test_models(self):

        # try getting model before retrieving
        with self.assertRaises(AssertionError):
            self.portfolio.get_model()

        with self.assertRaises(AssertionError):
            self.portfolio.set_models_weights({})

        # retrieve models
        market_tickers = ['^DJI']
        ranges = ['1mo', '6mo', '1y']
        self.assertFalse(self.portfolio.has_prices())
        self.assertFalse(self.portfolio.has_models())
        await self.portfolio.retrieve_models(market_tickers, ranges)
        self.assertFalse(self.portfolio.has_prices())
        self.assertTrue(self.portfolio.has_models())

        with self.assertRaises(AssertionError):
            await self.portfolio.retrieve_frontier(0, 0, False, True, True)

        from pyoptimum.model import Model

        model = self.portfolio.get_model()
        self.assertIsInstance(model, Model)

    async def test_models_with_prices(self):

        # try getting model before retrieving
        with self.assertRaises(AssertionError):
            self.portfolio.get_model()

        with self.assertRaises(AssertionError):
            self.portfolio.set_models_weights({})

        # retrieve models
        market_tickers = ['^DJI']
        ranges = ['1mo', '6mo', '1y']
        self.assertFalse(self.portfolio.has_prices())
        self.assertFalse(self.portfolio.has_models())
        await self.portfolio.retrieve_models(market_tickers, ranges, include_prices=True)
        self.assertTrue(self.portfolio.has_prices())
        self.assertTrue(self.portfolio.has_models())

        await self.portfolio.retrieve_frontier(0, 0, False, True, True)
        self.assertTrue(self.portfolio.has_frontier())

        from pyoptimum.model import Model

        model = self.portfolio.get_model()
        self.assertIsInstance(model, Model)

    async def test_frontier(self):

        # retrieve prices
        self.assertFalse(self.portfolio.has_prices())
        await self.portfolio.retrieve_prices()
        self.assertTrue(self.portfolio.has_prices())

        # retrieve models
        market_tickers = ['^DJI']
        ranges = ['1mo', '6mo', '1y']
        self.assertTrue(self.portfolio.has_prices())
        self.assertFalse(self.portfolio.has_models())
        await self.portfolio.retrieve_models(market_tickers, ranges)
        self.assertTrue(self.portfolio.has_prices())
        self.assertTrue(self.portfolio.has_models())

        # retrieve frontier
        await self.portfolio.retrieve_frontier(0, 100, False, True, True)
        self.assertTrue(self.portfolio.has_frontier())

        # retrieve unfeasible frontier
        with self.assertRaises(ValueError):
            await self.portfolio.retrieve_frontier(-100, 0, False, True, True)

        # make sure it gets invalidated
        self.assertFalse(self.portfolio.has_frontier())

        # set model weights
        self.assertDictEqual(self.portfolio.model_weights, {rg: 1/3 for rg in ranges})
        self.portfolio.set_models_weights({rg: v for rg, v in zip(ranges, [1,2,3])})
        self.assertDictEqual(self.portfolio.model_weights, {rg: v/6 for rg, v in zip(ranges, [1,2,3])})
        with self.assertRaises(AssertionError):
            self.portfolio.set_models_weights({})
        with self.assertRaises(AssertionError):
            self.portfolio.set_models_weights({rg: v for rg, v in zip(ranges, [1,-2,3])})


class TestPortfolioZeroShares(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        import pyoptimum
        from pyoptimum.portfolio import Portfolio

        self.portfolio_client = pyoptimum.AsyncClient(username=username,
                                                      password=password, api='optimize')
        self.model_client = pyoptimum.AsyncClient(username=username, password=password,
                                                  api='models')
        self.portfolio = Portfolio(self.portfolio_client, self.model_client)
        from pathlib import Path
        file = Path(__file__).parent / 'test_zero.csv'
        self.portfolio.import_csv(file)

    async def test_prices(self):

        self.assertEqual(self.portfolio.get_value(), 0.0)

        # retrieve prices
        self.assertFalse(self.portfolio.has_prices())
        await self.portfolio.retrieve_prices()
        self.assertTrue(self.portfolio.has_prices())

        self.assertEqual(self.portfolio.get_value(),
                         sum(self.portfolio.portfolio['value ($)']))
        self.assertEqual(self.portfolio.get_value(), 0.0)

        self.assertIn('close ($)', self.portfolio.portfolio)
        self.assertIn('value ($)', self.portfolio.portfolio)
        self.assertIn('value (%)', self.portfolio.portfolio)
        np.testing.assert_array_equal(self.portfolio.portfolio['value ($)'], 0.0)
        np.testing.assert_array_equal(self.portfolio.portfolio['value (%)'], 0.0)

        with self.assertRaises(AssertionError):
            await self.portfolio.retrieve_frontier(0, 0, False, True, True)

    async def test_frontier(self):

        # retrieve prices
        self.assertFalse(self.portfolio.has_prices())
        await self.portfolio.retrieve_prices()
        self.assertTrue(self.portfolio.has_prices())

        # retrieve models
        market_tickers = ['^DJI']
        ranges = ['1mo', '6mo', '1y']
        self.assertTrue(self.portfolio.has_prices())
        self.assertFalse(self.portfolio.has_models())
        await self.portfolio.retrieve_models(market_tickers, ranges)
        self.assertTrue(self.portfolio.has_prices())
        self.assertTrue(self.portfolio.has_models())

        # zero cashflow without short sales
        with self.assertRaises(ValueError) as e:
            await self.portfolio.retrieve_frontier(0, 100, False, True, True)
        self.assertTrue("Cashflow cannot be zero" in str(e.exception))

        # retrieve frontier
        await self.portfolio.retrieve_frontier(1000, 100, False, True, True)
        self.assertTrue(self.portfolio.has_frontier())
        x0 = self.portfolio.frontier['x']

        # retrieve frontier
        await self.portfolio.retrieve_frontier(-1000, 100, True, True, True)
        self.assertTrue(self.portfolio.has_frontier())
        np.testing.assert_array_almost_equal(self.portfolio.frontier['x'][0], -x0[0], 1e-4)

class TestPortfolioFunctions(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):

        import pyoptimum
        from pyoptimum.portfolio import Portfolio

        self.portfolio_client = pyoptimum.AsyncClient(username=username, password=password,api='optimize')
        self.model_client = pyoptimum.AsyncClient(username=username, password=password,api='models')
        self.portfolio = Portfolio(self.portfolio_client, self.model_client)
        from pathlib import Path
        file = Path(__file__).parent / 'test.csv'
        self.portfolio.import_csv(file)

        # retrieve models and price
        self.market_tickers = ['^DJI', '^RUT']
        self.ranges = ['1mo', '6mo', '1y']
        await self.portfolio.retrieve_models(self.market_tickers, self.ranges,
                                             include_prices=True)

    async def test_model_methods(self):

        from pyoptimum.model import Model

        weights = {
            '1mo': 3,
            '6mo': 1,
            '1y': 2
        }
        self.portfolio.set_models_weights(weights)
        weights = {
            '1mo': 3/6,
            '6mo': 1/6,
            '1y': 2/6
        }
        self.assertDictEqual(self.portfolio.model_weights, weights)

        self.assertEqual(self.portfolio.model_method, 'linear')
        model = self.portfolio.get_model()
        self.assertIsInstance(model, Model)

        # check model correctness
        np.testing.assert_array_almost_equal(1e10 * model.Q,
                                             1e10*((3/6) * self.portfolio.models['1mo'].Q +
                                                   (1/6) * self.portfolio.models['6mo'].Q +
                                                   (2/6) * self.portfolio.models['1y'].Q))
        np.testing.assert_array_almost_equal(model.F,
                                             (3/6) * self.portfolio.models['1mo'].F +
                                             (1/6) * self.portfolio.models['6mo'].F +
                                             (2/6) * self.portfolio.models['1y'].F)
        np.testing.assert_array_almost_equal(1e10 * model.D,
                                             1e10 * ((3/6) * self.portfolio.models['1mo'].D +
                                                     (1/6) * self.portfolio.models['6mo'].D +
                                                     (2/6) * self.portfolio.models['1y'].D))
        np.testing.assert_array_almost_equal(model.r,
                                             (3/6) * self.portfolio.models['1mo'].r +
                                             (1/6) * self.portfolio.models['6mo'].r +
                                             (2/6) * self.portfolio.models['1y'].r)

        self.portfolio.set_model_method('linear-fractional')
        self.assertEqual(self.portfolio.model_method, 'linear-fractional')
        model = self.portfolio.get_model()
        self.assertIsInstance(model, Model)

        # check model correctness
        np.testing.assert_array_almost_equal(1e10 * model.Q,
                                             1e10*((3/6) * self.portfolio.models['1mo'].Q +
                                                   (1/6) * self.portfolio.models['6mo'].Q +
                                                   (2/6) * self.portfolio.models['1y'].Q))
        np.testing.assert_array_almost_equal(model.F,
                                             (3/6) * self.portfolio.models['1mo'].F +
                                             (1/6) * self.portfolio.models['6mo'].F +
                                             (2/6) * self.portfolio.models['1y'].F)
        np.testing.assert_array_almost_equal(1e10 * model.Di,
                                             1e10 * ((3/6) * self.portfolio.models['1mo'].Di +
                                                     (1/6) * self.portfolio.models['6mo'].Di +
                                                     (2/6) * self.portfolio.models['1y'].Di))
        np.testing.assert_array_almost_equal(model.r,
                                             (3/6) * self.portfolio.models['1mo'].r +
                                             (1/6) * self.portfolio.models['6mo'].r +
                                             (2/6) * self.portfolio.models['1y'].r)

    async def test_apply_constraint(self):

        from pyoptimum.portfolio import LESS_THAN_OR_EQUAL, Portfolio

        tickers = ['MSFT']
        value = 1
        sign = LESS_THAN_OR_EQUAL
        for function in typing.get_args(Portfolio.ConstraintFunctionLiteral):
            for unit in typing.get_args(Portfolio.ConstraintUnitLiteral):
                self.portfolio.apply_constraint(tickers, function, sign, value, unit)
                if function == 'sales' or function == 'short sales':
                    self.assertTrue(np.isfinite(self.portfolio.portfolio.loc[tickers, 'lower']).all())
                    self.assertFalse(np.isfinite(self.portfolio.portfolio.loc[tickers, 'upper']).all())
                elif function == 'purchases' or function == 'holdings':
                    self.assertFalse(np.isfinite(self.portfolio.portfolio.loc[tickers, 'lower']).all())
                    self.assertTrue(np.isfinite(self.portfolio.portfolio.loc[tickers, 'upper']).all())
                self.portfolio.remove_constraints(tickers)

        tickers = ['MSFT', 'AAPL']
        value = 1
        sign = LESS_THAN_OR_EQUAL
        for function in typing.get_args(Portfolio.ConstraintFunctionLiteral):
            for unit in typing.get_args(Portfolio.ConstraintUnitLiteral):
                self.portfolio.apply_constraint(tickers, function, sign, value, unit)
                if function == 'purchases':
                    if unit == 'shares':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'upper'],
                                                      self.portfolio.portfolio.loc[tickers, 'shares'] + value)
                    elif unit == 'value':
                        np.testing.assert_array_equal(
                            self.portfolio.portfolio.loc[tickers, 'upper'],
                            self.portfolio.portfolio.loc[tickers, 'shares'] + value / self.portfolio.portfolio.loc[tickers, 'close ($)'])
                    elif unit == 'percent value':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'upper'],
                                                      (1 + value/100) * self.portfolio.portfolio.loc[tickers, 'shares'])
                elif function == 'sales':
                    if unit == 'shares':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'lower'],
                                                      self.portfolio.portfolio.loc[tickers, 'shares'] - value)
                    elif unit == 'value':
                        np.testing.assert_array_equal(
                            self.portfolio.portfolio.loc[tickers, 'lower'],
                            self.portfolio.portfolio.loc[tickers, 'shares'] - value / self.portfolio.portfolio.loc[tickers, 'close ($)'])
                    elif unit == 'percent value':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'lower'],
                                                      (1 - value/100) * self.portfolio.portfolio.loc[tickers, 'shares'])
                elif function == 'holdings':
                    if unit == 'shares':
                        self.assertTrue(np.all(self.portfolio.portfolio.loc[tickers, 'upper'] == value))
                    elif unit == 'value':
                        np.testing.assert_array_equal(
                            self.portfolio.portfolio.loc[tickers, 'upper'],
                            value / self.portfolio.portfolio.loc[tickers, 'close ($)'])
                    elif unit == 'percent value':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'upper'],
                                                      (value/100) * self.portfolio.portfolio.loc[tickers, 'shares'])
                elif function == 'short sales':
                    if unit == 'shares':
                        self.assertTrue(np.all(self.portfolio.portfolio.loc[tickers, 'lower'] == -value))
                    elif unit == 'value':
                        np.testing.assert_array_equal(
                            self.portfolio.portfolio.loc[tickers, 'lower'],
                            -value / self.portfolio.portfolio.loc[tickers, 'close ($)'])
                    elif unit == 'percent value':
                        np.testing.assert_array_equal(self.portfolio.portfolio.loc[tickers, 'lower'],
                                                      -(value/100) * self.portfolio.portfolio.loc[tickers, 'shares'])
                if function == 'sales' or function == 'short sales':
                    self.assertTrue(np.isfinite(self.portfolio.portfolio.loc[tickers, 'lower']).all())
                    self.assertFalse(np.isfinite(self.portfolio.portfolio.loc[tickers, 'upper']).all())
                elif function == 'purchases' or function == 'holdings':
                    self.assertFalse(np.isfinite(self.portfolio.portfolio.loc[tickers, 'lower']).all())
                    self.assertTrue(np.isfinite(self.portfolio.portfolio.loc[tickers, 'upper']).all())
                self.portfolio.remove_constraints(tickers)
